// Copyright Andrew Choi 1993, 2003.  Perl Artistic License

#import "MyView.h"

#include <math.h>
#include <unistd.h>

#include "boardstruc.h"
#include "boardprint.h"
#include "alphabeta.h"

#include "defs.h"

@implementation MyView

static int state = 0;  /* state of the user interface state machine */
static int pathLen;  /* number of marbles on current path */
static MarblePosType path[121];
static bool jumping;
static MarblePosType start;  /* start marble pos of a move */

static MarblePosType oppMoveFrom = -1;
static MarblePosType oppMoveTo = -1;

static NSColor *myColor;
static NSColor *oppColor;
static NSColor *myHighliteColor;
static NSColor *oppHighliteColor;

static int occupant[121];
static int occupantTemplate[121] = {
  0,
  0, 0,
  0, 0, 0,
  0, 0, 0, 0,
  1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1,
  1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1,
  1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1,
  1, 0, 0, 0, 0, 0, 0, 0, 0, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0,
  1, 0, 0, 0, 0, 0, 0, 0, 0, 1,
  1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1,
  1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1,
  1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1,
  0, 0, 0, 0,
  0, 0, 0,
  0, 0,
  0
};

static bool marked[121];
static MarblePosType returnPath[121];
static int returnPathLen;

/* Makes reuse of old Carbon code easier.  */
- (BOOL)isFlipped
{
  return YES;
}

- (id)initWithFrame:(NSRect)frameRect
{
  myColor = [[NSColor redColor] retain];
  oppColor = [[NSColor greenColor] retain];
  myHighliteColor = [[[myColor shadowWithLevel:0.6] colorWithAlphaComponent:0.5] retain];
  oppHighliteColor = [[[oppColor shadowWithLevel:0.6] colorWithAlphaComponent:0.5] retain];
  
  [super initWithFrame:frameRect];
  return self;
}

- (NSPoint)TranslateX:(int)x Y:(int)y
{
  float xx, yy;
  
  xx = x * OneOverRoot3 * VERTICAL_SKIP;
  yy = y * VERTICAL_SKIP;
  
  xx = xx + FRAME_WIDTH/2;
  yy = FRAME_HEIGHT - yy - VERTICAL_SKIP;
  
  return NSMakePoint(xx, yy);
}

- (void)DrawOneOrientation:(BoardStruc)posList
{
  int i, pos1, pos2;
  NSPoint p1, p2;
  
  for (i = 1; i < 16; i++) {
    pos1 = posList[i][0];
    pos2 = posList[i][posPerLine[i]-1];
    p1 = [self TranslateX:xPos[pos1] Y:yPos[pos1]];
    p2 = [self TranslateX:xPos[pos2] Y:yPos[pos2]];
    [[NSColor blackColor] set];
    [NSBezierPath setDefaultLineWidth:1.0];
    [NSBezierPath strokeLineFromPoint:p1 toPoint:p2];
  }
}

- (void)DrawOneHole:(MarblePosType)i
{
  NSRect holeRect;
  NSPoint p;
  
  p = [self TranslateX:xPos[i] Y:yPos[i]];
  holeRect.origin.x = p.x - HOLE_RADIUS;
  holeRect.origin.y = p.y - HOLE_RADIUS;
  holeRect.size.width = HOLE_RADIUS * 2;
  holeRect.size.height = HOLE_RADIUS * 2;

  switch (posColor[i]) {
  case LG:
    [[NSColor lightGrayColor] set];
    break;
  case MG:
    [[NSColor grayColor] set];
    break;
  case DG:
    [[NSColor darkGrayColor] set];
  }
  
  [[NSBezierPath bezierPathWithOvalInRect:holeRect] fill];
}

- (void)DrawPositions
{
  int i;
  
  for (i = 0; i < 121; i++)
    [self DrawOneHole:i];
}

- (void) DrawOneMarble:(MarblePosType)pos color:(NSColor *)marbleColor
{
  NSPoint p;
  NSRect marbleRect;
  NSBezierPath *marble;
  
  p = [self TranslateX:xPos[pos] Y:yPos[pos]];
  marbleRect.origin.x = p.x - MARBLE_RADIUS;
  marbleRect.origin.y = p.y - MARBLE_RADIUS;
  marbleRect.size.width = MARBLE_RADIUS * 2;
  marbleRect.size.height = MARBLE_RADIUS * 2;

  marble = [NSBezierPath bezierPathWithOvalInRect:marbleRect];
  [marbleColor set];
  [marble fill];
  [[NSColor blackColor] set];
  [marble setLineWidth:2.0];
  [marble stroke];
}

- (void)DrawMarbles
{
  int i;

  for (i = 0; i < 10; i++)
    [self DrawOneMarble:currBoard[0][i]
      color:myColor];
  
  for (i = 0; i < 10; i++)
    [self DrawOneMarble:currBoard[1][i]
      color:oppColor];
}

- (void)CircleMarble:(MarblePosType)pos color:pathColor
{
  int i = pos;
  NSPoint p;
  NSRect circleRect;
  NSBezierPath *circle;
  
  p = [self TranslateX:xPos[i] Y:yPos[i]];
  circleRect.origin.x = p.x - MARBLE_RADIUS;
  circleRect.origin.y = p.y - MARBLE_RADIUS;
  circleRect.size.width = MARBLE_RADIUS * 2;
  circleRect.size.height = MARBLE_RADIUS * 2;

  circle = [NSBezierPath bezierPathWithOvalInRect:circleRect];
  [pathColor set];
  [circle setLineWidth:4.0];
  [circle stroke];
}

- (void)DrawLollipopFrom:(MarblePosType)from to:(MarblePosType)to color:pathColor
{
  NSPoint start, end, stickStart, stickEnd;
  float dist;
  float dx, dy;
  NSRect shadedRect;
  NSBezierPath *circle;
	
  start = [self TranslateX:xPos[from] Y:yPos[from]];
  end = [self TranslateX:xPos[to] Y:yPos[to]];
  
  dx = end.x-start.x;
  dy = end.y-start.y;
  dist = sqrt(dx * dx + dy * dy);
  stickStart.x = (end.x-start.x) / dist * MARBLE_RADIUS + start.x;
  stickStart.y = (end.y-start.y) / dist * MARBLE_RADIUS + start.y;
  stickEnd.x = (start.x-end.x) / dist * MARBLE_RADIUS + end.x;
  stickEnd.y = (start.y-end.y) / dist * MARBLE_RADIUS + end.y;
    
  [pathColor set];
  [NSBezierPath setDefaultLineWidth:4.0];
  [NSBezierPath strokeLineFromPoint:stickStart toPoint:stickEnd];
  
  shadedRect.origin.x = end.x - MARBLE_RADIUS;
  shadedRect.origin.y = end.y - MARBLE_RADIUS;
  shadedRect.size.width = MARBLE_RADIUS * 2;
  shadedRect.size.height = MARBLE_RADIUS * 2;

  circle = [NSBezierPath bezierPathWithOvalInRect:shadedRect];
  [circle setLineWidth:4.0];
  [circle stroke];

  shadedRect.origin.x = end.x - LOLLIPOP_RADIUS;
  shadedRect.origin.y = end.y - LOLLIPOP_RADIUS;
  shadedRect.size.width = LOLLIPOP_RADIUS * 2;
  shadedRect.size.height = LOLLIPOP_RADIUS * 2;
  
  circle = [NSBezierPath bezierPathWithOvalInRect:shadedRect];
  [circle fill];
}

- (void)DrawPath:(PathType)path length:(int)pathLen color:(NSColor *)pathColor
{
  int i;
  
  [self CircleMarble:path[0] color:pathColor];

  for (i = 1; i < pathLen; i++)
    [self DrawLollipopFrom:path[i-1] to:path[i] color:pathColor];
  [self CircleMarble:path[pathLen-1] color:pathColor];
}

- (void)drawRect:(NSRect)rect
{
  [[NSColor whiteColor] set];
  NSRectFill([self frame]);
  
  [self DrawOneOrientation:horiz];
  [self DrawOneOrientation:slash];
  [self DrawOneOrientation:bSlash];
  
  [self DrawPositions];
  
  [self DrawMarbles];
  
  if (state == 1)
    [self DrawPath:path length:pathLen color:myHighliteColor];

  if (state == 0 && oppMoveTo > 0)
    [self CircleMarble:oppMoveTo color:oppHighliteColor];
}

- (MarblePosType)InvTranslate:(NSPoint) p
{
  float	xx, yy;
  int x, y;
  
  xx = p.x - FRAME_WIDTH/2;	   /* translate to pixel coord rel to pos 0 */
  yy = FRAME_HEIGHT - p.y - VERTICAL_SKIP;
  
  xx = xx / OneOverRoot3 / VERTICAL_SKIP;
  yy = yy / VERTICAL_SKIP;
  
  y = rint(yy);
  if (0 <= y && y <= 16) {
    x = rint((xx - hPosLeftCoord[y]) / 2);
    if (0 <= x && x < posPerLine[y])
      return(horiz[y][x]);
  }
  return(-1);
}

static bool OwnMarbleThere(MarblePosType pos)
{
  int i;
	
  for (i = 0; i < 10; i++)
    if (currBoard[0][i] == pos)
      return(true);
  return(false);
}

- (void)DrawSource:(MarblePosType)pos color:(NSColor *)marbleColor
{
  NSPoint p;
  NSRect shadedRect;
  
  p = [self TranslateX:xPos[pos] Y:yPos[pos]];

  shadedRect.origin.x = p.x - MARBLE_RADIUS;
  shadedRect.origin.y = p.y - MARBLE_RADIUS;
  shadedRect.size.width = MARBLE_RADIUS * 2;
  shadedRect.size.height = MARBLE_RADIUS * 2;

  [[NSColor yellowColor] set];
  [[NSBezierPath bezierPathWithOvalInRect:shadedRect] fill];
    
  [marbleColor set];
  [[NSBezierPath bezierPathWithOvalInRect:shadedRect] stroke];
}

- (void)DrawDestination:(MarblePosType)pos color:(NSColor *)marbleColor
{
  NSPoint p;
  NSRect shadedRect;
  
  p = [self TranslateX:xPos[pos] Y:yPos[pos]];
  
  shadedRect.origin.x = p.x - MARBLE_RADIUS;
  shadedRect.origin.y = p.y - MARBLE_RADIUS;
  shadedRect.size.width = MARBLE_RADIUS * 2;
  shadedRect.size.height = MARBLE_RADIUS * 2;

  [[marbleColor shadowWithLevel:0.5] set];
  [[NSBezierPath bezierPathWithOvalInRect:shadedRect] fill];
    
  [marbleColor set];
  [[NSBezierPath bezierPathWithOvalInRect:shadedRect] stroke];

  shadedRect.origin.x = p.x - BULLS_EYE_RADIUS;
  shadedRect.origin.y = p.y - BULLS_EYE_RADIUS;
  shadedRect.size.width = BULLS_EYE_RADIUS * 2;
  shadedRect.size.height = BULLS_EYE_RADIUS * 2;
    
  [[NSColor blackColor] set];
  [[NSBezierPath bezierPathWithOvalInRect:shadedRect] fill];
}

static void MakeMove(MarblePosType from, MarblePosType to)
{
  int i;
  
  for (i = 0; i < 10; i++)
    if (currBoard[0][i] == from) {
      currBoard[0][i] = to;
      break;
    }
}

static bool Jump(MarblePosType from, MarblePosType to, int depth);

static bool JumpAscendingJ(int i, int j, int to, MarblePosType posTable[17][16],
		    int depth)
{
  int k, l, bridgeLimit, landingPos;
  
  k = j+1;
  bridgeLimit = (posPerLine[i]-1 + j) / 2;
  while (k <= bridgeLimit && occupant[ posTable[i][k] ] < 2)
    k++;
  if (k <= bridgeLimit) {
    landingPos = k + (k - j);
    l = k+1;
    while (l <= landingPos && occupant[ posTable[i][l] ] < 2)
      l++;
    if (l > landingPos)
      if (! marked[ posTable[i][landingPos] ]) {
	marked[posTable[i][landingPos]] = true;
	if (posTable[i][landingPos] == to) {
	  returnPath[depth] = posTable[i][landingPos];
	  returnPathLen = depth+1;
	  return(true);
	}
	else {
	  returnPath[depth] = posTable[i][landingPos];
	  return(Jump(posTable[i][landingPos], to, depth+1));
	}
      }
  }
  return(false);
}

static bool JumpDescendingJ(int i, int j, int to, MarblePosType posTable[17][16],
		     int depth)
{
  int k, l, bridgeLimit, landingPos;
  
  k = j-1;
  bridgeLimit = (j+1) / 2;
  while (k >= bridgeLimit && occupant[ posTable[i][k] ] < 2)
    k--;
  if (k >= bridgeLimit) {
    landingPos = k - (j - k);
    l = k-1;
    while (l >= landingPos && occupant[ posTable[i][l] ] < 2)
      l--;
    if (l < landingPos)
      if (! marked[ posTable[i][landingPos] ]) {
	marked[posTable[i][landingPos]] = true;
	if (posTable[i][landingPos] == to) {
	  returnPath[depth] = posTable[i][landingPos];
	  returnPathLen = depth+1;
	  return(true);
	}
	else {
	  returnPath[depth] = posTable[i][landingPos];
	  return(Jump(posTable[i][landingPos], to, depth+1));
	}
      }
  }
  return(false);
}

static bool Jump(MarblePosType from, MarblePosType to, int depth)
{
  int i, j;
  
  i = lineInHoriz[from];
  j = posInLineHoriz[from];
  if (JumpAscendingJ(i, j, to, horiz, depth))
    return(true);
  if (JumpDescendingJ(i, j, to, horiz, depth))
    return(true);
  
  i = lineInSlash[from];
  j = posInLineSlash[from];
  if (JumpAscendingJ(i, j, to, slash, depth))
    return(true);
  if (JumpDescendingJ(i, j, to, slash, depth))
    return(true);
  
  i = lineInBSlash[from];
  j = posInLineBSlash[from];
  if (JumpAscendingJ(i, j, to, bSlash, depth))
    return(true);
  if (JumpDescendingJ(i, j, to, bSlash, depth))
    return(true);

  return false;
}

static void PathBetween(MarblePosType originally, MarblePosType from,
		 MarblePosType to, MarblePosType newPath[], int *newPathLen)
{
  int i, j;
  
  memcpy(&occupant, &occupantTemplate, sizeof(occupant));
  for (i = 0; i < 121; i++)
    marked[i] = false;
  marked[originally] = true;	/* never give starting position as move */
  
  for (i = 0; i < 10; i++)
    occupant[currBoard[0][i]] = 2;
  
  for (i = 0; i < 10; i++)
    occupant[currBoard[1][i]] = 3;

  /* once jump begins, position unoccupied, else jump over its
     nonexistent self */
  occupant[originally] = 0;
  
  *newPathLen = 0;
  
  i = lineInHoriz[from];
  j = posInLineHoriz[from];
  if (j > 0 && occupant[horiz[i][j-1]] == 0 && to == horiz[i][j-1])
    return;
  if (j<posPerLine[i]-1 && occupant[horiz[i][j+1]]==0 && to==horiz[i][j+1])
    return;
  
  i = lineInSlash[from];
  j = posInLineSlash[from];
  if (j > 0 && occupant[slash[i][j-1]] == 0 && to == slash[i][j-1])
    return;
  if (j<posPerLine[i]-1 && occupant[slash[i][j+1]]==0 && to==slash[i][j+1])
    return;
  
  i = lineInBSlash[from];
  j = posInLineBSlash[from];
  if (j > 0 && occupant[bSlash[i][j-1]] == 0 && to == bSlash[i][j-1])
    return;
  if (j<posPerLine[i]-1 && occupant[bSlash[i][j+1]]==0 && to==bSlash[i][j+1])
    return;
  
  if (Jump(from, to, 0)) {
    memcpy(newPath, returnPath, sizeof(returnPath));
    *newPathLen = returnPathLen;
    return;
  }
  
  *newPathLen = -1;
  return;
}

- (void)ComputeOppMove:(BoardType)oldBoard
{
  int i;
  
  for (i = 0; i < 10; i++)
    if (currBoard[1][i] != oldBoard[1][i]) {
      oppMoveFrom = oldBoard[1][i];
      oppMoveTo = currBoard[1][i];
    }
}

- (void)MakeOppMove
{
  BoardType oldBoard;

  memcpy(oldBoard, currBoard, sizeof(currBoard));
  MakeAMove(false);
  [self ComputeOppMove:oldBoard];
}

- (void)mouseDown:(NSEvent *)theEvent
{
  NSPoint clickLoc;
  MarblePosType clickedPos;
  int newPathLen;
  MarblePosType newPath[121];

  clickLoc = [self convertPoint:[theEvent locationInWindow] fromView:nil];
  clickedPos = [self InvTranslate:clickLoc];
  
  if (clickedPos >= 0) {
    switch (state) {
    case 0:
      if (OwnMarbleThere(clickedPos)) {
	start = clickedPos;
	path[0] = start;
	pathLen = 1;
	state = 1;
	jumping = false;
      }
      else {
        NSBeep();
      }
      break;
    case 1:
      if (clickedPos == start) {
	state = 0;
      }
      else if (clickedPos == path[pathLen-1]
	       && occupantTemplate[clickedPos] == 0) {
	MakeMove(start, path[pathLen-1]);
	state = 0;
        [self display];
	[self MakeOppMove];
        [self display];
      }
      else {
        int i;
	PathBetween(start, path[pathLen-1], clickedPos, newPath, &newPathLen);
	if (jumping && newPathLen > 0) {
	  for (i = 0; i < newPathLen; i++)
	    path[pathLen++] = newPath[i];
	}
	else {
	  PathBetween(start, start, clickedPos, newPath, &newPathLen);
	  if (newPathLen > 0) {
	    pathLen = 1;
	    for (i = 0; i < newPathLen; i++)
	      path[pathLen++] = newPath[i];
	    jumping = true;
	  }
	  else if (newPathLen == 0) {
	    path[1] = clickedPos;
	    pathLen = 2;
	    jumping = false;
	  }
	}
      }
      break;
    }
  }
  
  [self setNeedsDisplay:YES];
}

@end
