// Copyright Andrew Choi 1993, 2003.  Perl Artistic License

#include "boardstruc.h"
#include "static.h"
#include "movegen.h"
#include "alphabeta.h"
/* #include "drawboard.h" */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

int MinimaxAB(BoardType board, ConnectMapType connectMap, OccMapType occMap,
	      int depth, int useThresh, int passThresh);

int level = 5;

BoardType oldBoard;

BoardType bestPath[10];

#define TRACE_ALPHA_BETA 0

void MakeAMove(bool myTurn)
{
  int moveValue, k;
  BoardType opponentBoard;
  ConnectMapType connectMap;
  OccMapType occMap;
	
  if (myTurn) {
    BuildConnectMap(currBoard, connectMap, occMap);
    moveValue = MinimaxAB(currBoard, connectMap, occMap, 0, 9999, -9999);
    for (k = 0; k < 10; k++)
      currBoard[0][k] = bestPath[0][0][k];	
    currBoard[1][k] = bestPath[0][1][k];	
  }
  else {
    for (k = 0; k < 10; k++) {
      opponentBoard[0][k] = 120 - currBoard[1][k];
      opponentBoard[1][k] = 120 - currBoard[0][k];
    }
    BuildConnectMap(opponentBoard, connectMap, occMap);
    moveValue = MinimaxAB(opponentBoard, connectMap, occMap, 0, 9999, -9999);
    for (k = 0; k < 10; k++) {
      currBoard[0][k] = 120 - bestPath[0][1][k];
      currBoard[1][k] = 120 - bestPath[0][0][k];
    }
  }
  /* SetCursor(&qd.arrow); */
}	

bool DeepEnough(int depth)
{
  return(depth >= level);
}

BoardType bestPath[10];
bool Toggle = true;

int MinimaxAB(BoardType board, ConnectMapType connectMap, OccMapType occMap,
	      int depth, int useThresh, int passThresh)
{
  MoveList successors;
  int i, j, numSuccessors, newValue;
  BoardType succ;
  ConnectMapType succCM;
  OccMapType succOM;
	
  /* if (depth <=2) BlinkAlive(); */

  if (DeepEnough(depth)) {
    i = StaticEval(board, depth);
#if TRACE_ALPHA_BETA
    for (j = 0; j < depth; j++)
      printf("  ");
    printf("%d (level %d)\n", i, depth);
#endif
    return(i);
  }

  if (depth % 2 == 0)
    SelfMoveGen(board, connectMap, occMap, &successors, &numSuccessors);
  else
    OppMoveGen(board, connectMap, occMap, &successors, &numSuccessors);
  
  /* numSuccessors = GetPtrSize(successors)/sizeof(MoveType); */
  for (i = 0; i < numSuccessors; i++) {
    /* BlockMove(board, succ, sizeof(BoardType));
    BlockMove(connectMap, succCM, sizeof(ConnectMapType));
    BlockMove(occMap, succOM, sizeof(OccMapType)); */
    memcpy(succ, board, sizeof(BoardType));
    memcpy(succCM, connectMap, sizeof(ConnectMapType));
    memcpy(succOM, occMap, sizeof(OccMapType));
    CMDelete(successors[i].from, succCM, succOM);
    CMAdd(successors[i].to, succCM, succOM);
    for (j = 0; j < 10; j++) {		/* this could be done better */
      if (succ[0][j] == successors[i].from)
	succ[0][j] = successors[i].to;
      if (succ[1][j] == successors[i].from)
	succ[1][j] = successors[i].to;
    }
    newValue = -MinimaxAB(succ, succCM, succOM, depth+1, -passThresh,
			  -useThresh);
    if (newValue > passThresh) {
      passThresh = newValue;
      /* BlockMove(&succ, &(bestPath[depth]), sizeof(succ)); */
      memcpy(&(bestPath[depth]), &succ, sizeof(succ));
    }
    if (passThresh >= useThresh) {
#if TRACE_ALPHA_BETA
      for (j = 0; j < depth; j++)
	printf("  ");
      printf("%d (level %d)\n", passThresh, depth);
#endif
      /* DisposPtr(successors); */
      free(successors);
      return(passThresh);
    }
  }
  
  /* DisposPtr(successors); */
  free(successors);
#if TRACE_ALPHA_BETA
  for (j = 0; j < depth; j++)
    printf("  ");
  printf("%d (level %d)\n", passThresh, depth);
#endif
  return(passThresh);
}

/* BlinkAlive()
{	
  TurnBoingCursor();
}
*/
