// Copyright Andrew Choi 1993, 2003.  Perl Artistic License

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "boardstruc.h"
#include "movegen.h"

void AddPos(MarblePosType start, MarblePosType dest);
void RecursiveJump(MarblePosType originalStart, MarblePosType start,
		   BoardType board, ConnectMapType cM, OccMapType oM);
void MoveGen1Marble(MarblePosType start, BoardType board, ConnectMapType cM,
		    OccMapType oM);
void SelectSelfMoves(void);
void SelectOppMoves(void);
void Add(MarblePosType pos, ConnectMapType cM, OccMapType oM,
	MarblePosType posTab[17][16], int lineInTab[], int posInLineTab[],
	int plus, int minus);
void Delete(MarblePosType pos, ConnectMapType cM, OccMapType oM,
	MarblePosType posTab[17][16], int lineInTab[], int posInLineTab[],
	int plus, int minus);

#if 0
#define	rNoMemSelfAlert	129	/* out of memory in SelfMoveGen */
#define	rNoMemOppAlert	130	/* out of memory in OppMoveGen */

static void AlertUser(int alertID)
{
  short itemHit;

  SetCursor(&qd.arrow);
  itemHit = Alert(alertID, nil);
  ExitToShell();
}
#endif

void BuildConnectMap(BoardType board, ConnectMapType cM, OccMapType oM)
{
  int i, j;
	
  for (i = 0; i < 121; i++)
    oM[i] = false;

  for (i = 0; i < 10; i++) {
    oM[board[0][i]] = true;
    oM[board[1][i]] = true;
  }
	
  for (i = 0; i < 121; i++)
    for (j = 0; j < 6; j++)
      cM[i][j] = -1;
		
  for (i = 0; i < 2; i++)
    for (j = 0; j < 10; j++)
      CMAdd(board[i][j], cM, oM);
}

void CMAdd(MarblePosType pos, ConnectMapType cM, OccMapType oM)
{
  oM[pos] = true;

  Add(pos, cM, oM, horiz, lineInHoriz, posInLineHoriz, HorizPlus, HorizMinus);
  Add(pos, cM, oM, slash, lineInSlash, posInLineSlash, SlashPlus, SlashMinus);
  Add(pos, cM, oM, bSlash, lineInBSlash, posInLineBSlash, BSlashPlus,
      BSlashMinus);
}

void Add(MarblePosType pos, ConnectMapType cM, OccMapType oM,
	 MarblePosType posTab[17][16], int lineInTab[], int posInLineTab[],
	 int plus, int minus)
{
  int i, j, k, l;
	
  i = lineInTab[pos];
  j = posInLineTab[pos];
	
  k = j - 1;
  l = j + 1;
  while (k>=0 && l<posPerLine[i] && !oM[posTab[i][k]] && !oM[posTab[i][l]]) {
    cM[posTab[i][k]][plus] = posTab[i][l];
    cM[posTab[i][l]][minus] = posTab[i][k];
    k--;
    l++;
  }
	
  if (k >= 0 && l < posPerLine[i] && oM[posTab[i][k]])
    cM[posTab[i][k]][plus] = posTab[i][l];
  if (k >= 0 && l < posPerLine[i] && oM[posTab[i][l]])
    cM[posTab[i][l]][minus] = posTab[i][k];
}

void CMDelete(MarblePosType pos, ConnectMapType cM, OccMapType oM)
{
  oM[pos] = false;

  Delete(pos, cM, oM, horiz, lineInHoriz, posInLineHoriz, HorizPlus,
	 HorizMinus);
  Delete(pos, cM, oM, slash, lineInSlash, posInLineSlash, SlashPlus,
	 SlashMinus);
  Delete(pos, cM, oM, bSlash, lineInBSlash, posInLineBSlash, BSlashPlus,
	 BSlashMinus);
}

void Delete(MarblePosType pos, ConnectMapType cM, OccMapType oM,
	    MarblePosType posTab[17][16], int lineInTab[], int posInLineTab[],
	    int plus, int minus)
{
  int i, j, k, l;
	
  i = lineInTab[pos];
  j = posInLineTab[pos];
	
  k = j - 1;
  l = j + 1;
  while (k>=0 && l<posPerLine[i] && !oM[posTab[i][k]] && !oM[posTab[i][l]]) {
    cM[posTab[i][k]][plus] = -1;
    cM[posTab[i][l]][minus] = -1;
    k--;
    l++;
  }
	
  if (k >= 0 && l < posPerLine[i] && oM[posTab[i][k]])
    cM[posTab[i][k]][plus] = -1;
  if (k >= 0 && l < posPerLine[i] && oM[posTab[i][l]])
    cM[posTab[i][l]][minus] = -1;
	
  while (k >= 0 && !oM[posTab[i][k]])
    k--;
  if (k >= 0)
    Add(posTab[i][k], cM, oM, posTab, lineInTab, posInLineTab, plus, minus);

  while (l < posPerLine[i] && !oM[posTab[i][l]])
    l++;
  if (l < posPerLine[i])
    Add(posTab[i][l], cM, oM, posTab, lineInTab, posInLineTab, plus, minus);
}

MoveType moves[1000];
int numMoves;
	
void SelfMoveGen(BoardType board, ConnectMapType cM, OccMapType oM,
		 MoveList *result, int *resNumMoves)
{
  int i;
	
  numMoves = 0;
	
  for (i = 0; i < 10; i++)
    MoveGen1Marble(board[0][i], board, cM, oM);
	
  SelectSelfMoves();
		
  /* *result = (MoveList) NewPtr(numMoves * sizeof(MoveType)); */
  *result = (MoveList) malloc(numMoves * sizeof(MoveType));
  /* if (*result == NULL)
     AlertUser(rNoMemSelfAlert); */
  assert(*result);
  /* BlockMove(moves, *result, numMoves * sizeof(MoveType)); */
  memcpy(*result, moves, numMoves * sizeof(MoveType));
  *resNumMoves = numMoves;
}

void OppMoveGen(BoardType board, ConnectMapType cM, OccMapType oM,
		MoveList *result, int *resNumMoves)
{
  int i;

  numMoves = 0;
	
  for (i = 0; i < 10; i++)
    MoveGen1Marble(board[1][i], board, cM, oM);
	
  SelectOppMoves();

  /* *result = (MoveList) NewPtr(numMoves * sizeof(MoveType)); */
  *result = (MoveList) malloc(numMoves * sizeof(MoveType));
  /* if (*result == NULL)
     AlertUser(rNoMemOppAlert); */
  assert(*result);
  /* BlockMove(moves, *result, numMoves * sizeof(MoveType)); */
  memcpy(*result, moves, numMoves * sizeof(MoveType));
  *resNumMoves = numMoves;
}

/* 0=free, 1=neutral zone, 2=own, 3=opponent */
int occKind[121];
int occKindTemplate[121] = {
  0,
  0, 0,
  0, 0, 0,
  0, 0, 0, 0,
  1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1,
  1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1,
  1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1,
  1, 0, 0, 0, 0, 0, 0, 0, 0, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0,
  1, 0, 0, 0, 0, 0, 0, 0, 0, 1,
  1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1,
  1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1,
  1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1,
  0, 0, 0, 0,
  0, 0, 0,
  0, 0,
  0
};

bool visited[121];

void MoveGen1Marble(MarblePosType start, BoardType board, ConnectMapType cM,
		    OccMapType oM)
{
  int i, j;
  ConnectMapType cM2;
  OccMapType oM2;

  for (i = 0; i < 121; i++)
    visited[i] = false;
  visited[start] = true;	/* never give starting position as move */

  /* BlockMove((Ptr) &occKindTemplate, (Ptr) &occKind, sizeof(occKind)); */
  memcpy(&occKind, &occKindTemplate, sizeof(occKind));

  for (i = 0; i < 10; i++)
    occKind[board[0][i]] = 2;

  for (i = 0; i < 10; i++)
    occKind[board[1][i]] = 3;

  /* once jump begins, position unoccupied, else jump over its
     nonexistent self */
  occKind[start] = 0;
	
  /* BlockMove(cM, cM2, sizeof(ConnectMapType));
     BlockMove(oM, oM2, sizeof(OccMapType)); */
  memcpy(cM2, cM, sizeof(ConnectMapType));
  memcpy(oM2, oM, sizeof(OccMapType));
  CMDelete(start, cM2, oM2);
  
  RecursiveJump(start, start, board, cM2, oM2);

  i = lineInHoriz[start];
  j = posInLineHoriz[start];
  if (j > 0 && occKind[ horiz[i][j-1] ] == 0)
    AddPos(start, horiz[i][j-1]);
  if (j < posPerLine[i]-1 && occKind[ horiz[i][j+1] ] == 0)
    AddPos(start, horiz[i][j+1]);
	
  i = lineInSlash[start];
  j = posInLineSlash[start];
  if (j > 0 && occKind[ slash[i][j-1] ] == 0)
    AddPos(start, slash[i][j-1]);
  if (j < posPerLine[i]-1 && occKind[ slash[i][j+1] ] == 0)
    AddPos(start, slash[i][j+1]);
	
  i = lineInBSlash[start];
  j = posInLineBSlash[start];
  if (j > 0 && occKind[ bSlash[i][j-1] ] == 0)
    AddPos(start, bSlash[i][j-1]);
  if (j < posPerLine[i]-1 && occKind[ bSlash[i][j+1] ] == 0)
    AddPos(start, bSlash[i][j+1]);
}

void RecursiveJump(MarblePosType originalStart, MarblePosType start,
		   BoardType board, ConnectMapType cM, OccMapType oM)
{
  int i;
  
  for (i = 0; i < 6; i++)
    if (cM[start][i] != -1 && !visited[cM[start][i]]
	&& occKind[cM[start][i]] < 2) {
      AddPos(originalStart, cM[start][i]);
      RecursiveJump(originalStart, cM[start][i], board, cM, oM);
    }
			
}

void AddPos(MarblePosType start, MarblePosType dest)
{
  visited[dest] = true;
  if (occKind[dest] == 0) {
    moves[numMoves].from = start;
    moves[numMoves].to = dest;
    numMoves++;
  }
}

typedef struct {
  int from;
  int to;
  int moveVal;
} MVType;
	
MVType mV[1000];

int compar1(const void *e1, const void *e2)
{
  return(((MVType *) e2)->moveVal  - ((MVType *) e1)->moveVal);
}

void SelectSelfMoves(void)	/* do it in place "moves"! */
{
  int j;

  for (j = 0; j < numMoves; j++) {
    mV[j].from = moves[j].from;
    mV[j].to = moves[j].to;
    mV[j].moveVal = lineInHoriz[moves[j].to]-lineInHoriz[moves[j].from];
  }

  qsort(mV, numMoves, sizeof(MVType), compar1);
	
  for (j = 0; j < numMoves; j++) {
    moves[j].from = mV[j].from;
    moves[j].to = mV[j].to;
  }
}

int compar2(const void *e1, const void *e2)
{
  return(((MVType *) e1)->moveVal  - ((MVType *) e2)->moveVal);
}

void SelectOppMoves(void)	/* do it in place "moves"! */
{
  int j;

  for (j = 0; j < numMoves; j++) {
    mV[j].from = moves[j].from;
    mV[j].to = moves[j].to;
    mV[j].moveVal = lineInHoriz[moves[j].to]-lineInHoriz[moves[j].from];
  }

  qsort(mV, numMoves, sizeof(MVType), compar2);
	
  for (j = 0; j < numMoves; j++) {
    moves[j].from = mV[j].from;
    moves[j].to = mV[j].to;
  }
}
